% Perform illumination cone based 3D face reconstruction using ARfM
%
% Jiansheng Chen, Chang Yang, Yu Deng, Gang Zhang and Guangda Su, Exploring Facial 
% Asymmetry Using Optical Flow, IEEE Signal Processing Letters, Vol. 21, No. 7, pp. 792
% -795, 2014
%
%01-04-14, Jiansheng Chen

function arfm3D(I0, I2)

    % convert the image to gray scale
    try
        I0 = rgb2gray(I0);
        I2 = rgb2gray(I2);
    catch exception
    end
    I0 = im2double(I0);
    I1 = im2double(I2);
    
    % calculate the reference face image
    [H,W] = size(I0);
    I0_left = mean(mean(I0(:,1:round(W/2))));
    I0_right = mean(mean(I0(:,round(W/2)+1:end)));
    I1_left = mean(mean(I1(:,1:round(W/2))));
    I1_right = mean(mean(I1(:,round(W/2)+1:end)));
    A = [1 1; I0_left-I0_right I1_left-I1_right];
    coef = A\[1;0];
    I_ref = I0*coef(1)+I1*coef(2);
    
    % calculate the OF-Face of the reference image
    alpha = 0.01; ratio = 0.75;
    minWidth = 20; nOuterFPIterations = 7;
    nInnerFPIterations = 1;  nSORIterations = 30;
    para = [alpha,ratio,minWidth,nOuterFPIterations,nInnerFPIterations,nSORIterations];    
    I_ref_flip = fliplr(I_ref);
    [vx,vy,test] = Coarse2FineTwoFrames(I_ref, I_ref_flip, para);
    flow(:,:,1) = vx;  flow(:,:,2) = vy;
    imgFlow = flowToColor(flow);
    
    % display the OF-Face
    figure('Name', 'Refrence Face and OF-Face');
    subplot(1,2,1); imshow(I_ref);
    subplot(1,2,2); imshow(imgFlow);
    
    % generate ARfM of input images
    I0_ARfM = WarpFrameGivenFlow(fliplr(I0), fliplr(I0), vx, vy);
    I1_ARfM = WarpFrameGivenFlow(fliplr(I1), fliplr(I1), vx, vy);
    
    % prepare data for 3D face reconstruction
    D = zeros(numel(I0), 4);
    D(:,1) = I0(:); D(:,2) = I1(:); 
    D(:,3) = I0_ARfM(:); D(:,4) = I1_ARfM(:);
    D(D>1) = 1; D(D<0) = 0;
    
    % illumination cone base 3D face reconstruction
    disp('Start 3D face reconstruction ...');
    [~, ~, albedo, z] = constructIlluminationCone(D, W, H);
    
    % display 3D reconstruction results
    figure;
    h = surf(1:size(z,2),size(z,1):-1:1,z, albedo);
    colormap 'gray';axis equal;
    axis off; set(h,'EdgeColor','none');
    set(h,'AmbientStrength',0.6); set(h,'SpecularStrength',0); view(0,90);
    src = [0 0 1.5];
    light('Position',src,'Style','infinite');
    view(77, 48);
    
    % finalize
    pause; disp('press any key to continue ...');
    close all;
end