% following the paper "From Few to Many: Illumination Cone Models for Face
% Recognition under Variable Lighting and Pose"
function [B, S, albedo,z] = constructIlluminationCone(X, imWidth, imHeight)
    
    tstart = tic;
    
    %*********************************%
    % mark the saturation and shadows %
    %*********************************%
    % mark the saturation
    W = (X<=0.96).*(X>=0.01);
    
    %As an initial estimate of the albedo, we use the average of the
    %training images.
    albedo = mean(X,2);
%     imshow(reshape(albedo, 192, 168));
    W = W.*( (X./((albedo+eps)*ones(1, size(X,2))))>=0.618 );
%     imshow(reshape(W, 192, 168));

    %********************************************************************%
    % Without doing any row or column permutations, sift out all of the  %
    % full rows (with no missing measurements) of matrix X to form a full%
    % submatrix Xcap                                                     %
    %********************************************************************%
    Wmask = logical(prod(W,2));
    Xcap = X(Wmask, :);
    
    %**************************************************************%
    % Perform SVD on Xcap to find an initial estimate of matrix S  %
    %**************************************************************%
%     [U,M,V] = svd(Xcap, 0);
    [U,M,V] = svd(X, 0);
    Sall = M*V';
    S = zeros(3, size(Sall, 2));
    S(1, :) = Sall(2, :);
    S(2, :) = Sall(3, :);
    S(3, :) = Sall(1, :);
    if sum(U(:,1))>0
        S = -S;
    end
    clear U M V;
    
        
    [B, S, albedo, z] = step456789(X, S, W, albedo, imHeight, imWidth);
    epsilon = 1000000;
    iter = 0;
    Rerror_min = 1000000;
    
    while epsilon>0.002 && iter<20
        iter = iter+1;
        [B_next, S_next, albedo_next, z_next] = step456789(X, S, W, albedo, imHeight, imWidth);
        fprintf(1, 'Iteration %d\n', iter);
        Rerror = sum(sum(abs(X(Wmask, :)-B_next(Wmask, :)*S_next)))/sum(Wmask);
        fprintf(1, 'Reconstruction error = %f\n', Rerror);
        epsilon = sum(sum(abs(S_next-S)))/sum(abs(S(:)));
        fprintf(1, 'Covergence epsilon = %f\n', epsilon);
        B = B_next; S = S_next; albedo = albedo_next; z = z_next;
        if Rerror < Rerror_min
            B_min = B; S_min = S; albedo_min = albedo; z_min = z;
            Rerror_min = Rerror;
        end
        pause(0.1);
    end
    
    B = B_min; S = S_min; albedo = albedo_min; z = z_min;
    
    [z,~,~,~] = AffineTransformation(reshape(-B(:,1)./B(:,3), imHeight, imWidth),reshape(-B(:,2)./B(:,3), imHeight, imWidth));
    albedo = sqrt(B(:,1).^2 + B(:,2).^2 + B(:,3).^2);
    albedo(albedo>mean(albedo)+3*std(albedo)) = mean(albedo)+3*std(albedo);
    albedo = reshape(mean(X,2), imHeight, imWidth);

    ttime = toc(tstart);
    
    fprintf(1, '\nDone in %0.1f seconds within %d iterations, reconstruction error = %f.\n\n', ttime, iter, Rerror_min);

    return;
    
    function [B, S, albedo, z] = step456789(X, S, W, albedo, imHeight, imWidth)
    
    %***************************************************************************%
    % Find the vectors bj (the rows of B) by performing the minimization in (7)%
    %***************************************************************************%
    Bstar = ((S*S')\(S*X'))';
    
    if sum(Bstar(:,3))>0
        Bstar = -Bstar;
    end
    
    idx = find(logical(prod(W,2))==0);
    W = logical(W);
    
    for j = 1:length(idx)
        Xj = X(idx(j),:)';
        H = S(:, W(idx(j),:)')*S(:, W(idx(j),:)')';
        f = -S(:, W(idx(j),:)')*Xj(W(idx(j),:)');
        if rcond(H)<1e-5  %for sum(Wj)<3, ill conditioned cases
            if idx(j)==1
                bj = [0; 0; -1];
            else
                bj = Bstar(idx(j)-1, :)';
            end
         else
             bj = -H\f;
         end            

        Bstar(idx(j), :) = bj';
    end
    
    % should use constraints that the third column is nonpositive
    idx = find(Bstar(:,3)>0);
    C = [0 0 1];
    for j = 1:length(idx)
        Xj = X(idx(j),:)';
        H = S(1:2, W(idx(j),:)')*S(1:2, W(idx(j),:)')';
        f = -S(1:2, W(idx(j),:)')*Xj(W(idx(j),:)');
        if rcond(H)<1e-5  %for sum(Wj)<3, ill conditioned cases
            if idx(j)==1
                bj = [0; 0; -1];
            else
                bj = Bstar(idx(j)-1, :)';
            end
         else
             bj = [-H\f; 0];
         end            
        Bstar(j, :) = bj;
    end
    
    clear H  Xj  bj f j;
    
    %***************************************************************************%
    % Estimate a possibly nonintegrable set of partial derivatives z*_x z*_y    %
    %***************************************************************************%
    tmp = Bstar./((albedo+eps)*ones(1,3));
    
    % handling some possible divide by zero problem and negtive albedo
    % problem
    idx = find(tmp(:,3)>=0);
    if length(idx) >= size(tmp,1)
        error('Invalid B data!');
    elseif length(idx)>=1
        if idx(1)==1
            j = 2;
            while tmp(j, 3)>=0
                j = j+1;
            end
            tmp(1, :) = tmp(j, :);
        end
        idx = find(tmp(:,3)>=0);
        %use previous data to approx
        for i=1:length(idx)
            j = idx(i)-1;
            while tmp(j, 3)>=0
                j = j-1;
            end
            tmp(idx(i), :) = tmp(j, :);
        end
    end

    Zx_star = reshape(-tmp(:,1)./tmp(:,3), imHeight, imWidth);
    Zy_star = reshape(-tmp(:,2)./tmp(:,3), imHeight, imWidth);

    clear tmp i j idx;

    %***************************************************************************%
    % Estimate (as functions of c(w) a set of integrable partial derivatives  %
    %***************************************************************************%
    [z,~,~,~] = AffineTransformation(Zx_star,Zy_star);
    [Zx, Zy] = calculate_gradients(z, 0,  0);
%     Zx = removeOutlier(Zx);
%     Zy = removeOutlier(Zy);

    
    %*************************************************%
    % Update the albedo by least-squares minimization %
    %*************************************************%
    A = [reshape(Zx, numel(Zx), 1) reshape(Zy, numel(Zy), 1) -ones(numel(Zx), 1)];
    A = A./(sqrt(sum(A.^2, 2))*ones(1, 3));
    AS = (A*S).*W;
        
    albedo = sum(AS.*X, 2)./sum(AS.^2, 2);
    % if for a pixel all W=0, than got NaN for albedo, solve this by
    % assuming albedo continuity
    albedo2 = reshape(albedo, imHeight, imWidth);
    [row, col] = find(isnan(albedo2));
    for k = 1:length(row)
        block = albedo2(max(1, row(k)-2):min(imHeight, row(k)+2), max(1, col(k)-2):min(imWidth, col(k)+2));
        if sum(isnan(block)) <=0
            albedo2(row(k),  col(k)) = 0.2;
        else
            albedo2(row(k),  col(k)) = mean(mean(block(~isnan(block))));
        end
    end
    albedo = reshape(albedo2, numel(albedo2), 1);

    
    %***************************************************************************%
    % Construct B by using the newly calculated albedo the partial derivatives %
    %***************************************************************************%    
    B = (albedo*ones(1,3)).*A;
    
    %***************************************************************************%
    % Update each of the light source directions and strengths si independently %
    % using least-squares minimization and the newly constructed B             %
    %***************************************************************************%     
    H = 2*(B'*B);
    for i = 1:size(S,2)
        f = (-2*X(:,i)'*B)';
        S(:,i) = -H\f;
    end
    
    albedo = sqrt(B(:,1).^2 + B(:,2).^2 + B(:,3).^2); 
    [z,~,~,~] = AffineTransformation(reshape(-B(:,1)./B(:,3), imHeight, imWidth),reshape(-B(:,2)./B(:,3), imHeight, imWidth));
    
    return;
               
